#
#   Foosic / libFooID analysis
#   Copyright (C) 2012 Gian-Carlo Pascutto <gcp@sjeng.org>

from __future__ import print_function
from struct import Struct

FRAMES_PER_FP = 87
MINLENGTH = 20
MATCH_SCORE = 70

class FooID:
    def __init__(self, buffer=None, songname=None):
        self.version = 0
        self.length = 0
        self.avg_fit = 0
        self.avg_dom = 0    
        self.fits = [None] * FRAMES_PER_FP * 16
        self.doms = []
        self.songname = None
        if buffer:
            self.decode_from(buffer)
    def __str__(self):        
        res = ("Version %d, length %d, avg_fit %d, avg_dom %d\n"
               % (self.version, self.length, self.avg_fit, self.avg_dom))
        return res
    def unpack_fits(self, buffer):    
        i = 0
        for bytebuff in buffer:
            self.fits[i]   = ((bytebuff >> 6) & 0x3)
            self.fits[i+1] = ((bytebuff >> 4) & 0x3)
            self.fits[i+2] = ((bytebuff >> 2) & 0x3)
            self.fits[i+3] = ((bytebuff >> 0) & 0x3)
            i += 4
        assert len(self.fits) == (FRAMES_PER_FP * 16)        
    def unpack_doms(self, buffer):    
        for offset in xrange(0, len(buffer), 3):
            self.doms.extend(((buffer[offset+0] >> 2),
                            (((buffer[offset+0] & 0x3) << 4) | (buffer[offset+1] >> 4)),
                            (((buffer[offset+1] & 0xF) << 2) | (buffer[offset+2] >> 6)),
                             ((buffer[offset+2] & 0x3F))))
        # last 6 bits are random data, there's only 87 doms
        self.doms = self.doms[0:FRAMES_PER_FP]
        assert len(self.doms) == FRAMES_PER_FP                
    def decode_from(self, fp):
        header = Struct("<hihh")    
        self.version, self.length, self.avg_fit, self.avg_dom =\
            header.unpack_from(fp, 0)    
        fit_struct = Struct("<348B")
        fit_data = fit_struct.unpack_from(fp, header.size)
        dom_struct = Struct("<66B")
        dom_data = dom_struct.unpack_from(fp, header.size + fit_struct.size)
        self.unpack_fits(fit_data)
        self.unpack_doms(dom_data)                   
    def fp_match(self, other):
        fit1, fit2 = self.fits, other.fits
        dom1, dom2 = self.doms, other.doms
        # sample rate / FFT size
        ratio = 8000.0 / 8192.0
        # frames in this fingerprint
        frames = min(FRAMES_PER_FP, round((self.length * ratio) / 100.0))
        # maximum error possible given length
        maxerror = frames * ((9 * 16) + (0.25 * 63))
        error_fit = error_dom = 0
        for i in xrange(len(fit1)):
            error_fit += (fit1[i] - fit2[i])**2
        for i in xrange(len(dom1)):
            error_dom += abs(dom1[i] - dom2[i])
        error = error_fit + (error_dom / 4.0)
        perc = 1.0 - (error / maxerror)
        conf = (perc - 0.5) * 2 * 100
        conf = max(conf, 0.0)
        return conf
    def get_songname(self):
        return self.songname
